
from enum import Enum
from typing import Optional


class MudraBLEServicesUUID(Enum):
    COMMAND_SERVICE = "0000fff0-0000-1000-8000-00805f9b34fb"
    BATTERY_SERVICE = "0000180f-0000-1000-8000-00805f9b34fb"
    DFU_SERVICE = "0000fe59-0000-1000-8000-00805f9b34fb"
    INFORMATION_SERVICE = "0000180a-0000-1000-8000-00805f9b34fb"

    @property
    def value_str(self) -> str:
        return self.value

    @staticmethod
    def from_value(value: str) -> Optional["MudraBLEServicesUUID"]:
        for item in MudraBLEServicesUUID:
            if item.value == value:
                return item
        return None

class MudraCharacteristicUUID(Enum):
    SNC_CHARACTERISTIC = "0000fff4-0000-1000-8000-00805f9b34fb"
    IMU_CHARACTERISTIC = "0000fff5-0000-1000-8000-00805f9b34fb"
    COMMAND_CHARACTERISTIC = "0000fff1-0000-1000-8000-00805f9b34fb"
    LOGGING_CHARACTERISTIC = "0000fff2-0000-1000-8000-00805f9b34fb"
    MESSAGE_CHARACTERISTIC = "0000fff6-0000-1000-8000-00805f9b34fb"
    BATTERY_CHARACTERISTIC = "00002a19-0000-1000-8000-00805f9b34fb"
    FIRMWARE_CHARACTERISTIC = "00002a26-0000-1000-8000-00805f9b34fb"
    SERIAL_RIGHT_PART_CHARACTERISTIC = "00002a25-0000-1000-8000-00805f9b34fb"
    SERIAL_LEFT_PART_CHARACTERISTIC = "00002a27-0000-1000-8000-00805f9b34fb"
    DFU_WITHOUT_BONDS = "8ec90003-f315-4f60-9fb8-838830daea50"
    CHARGING_STATE = "00002a1a-0000-1000-8000-00805f9b34fb"

    @property
    def value_str(self) -> str:
        return self.value

    @staticmethod
    def from_value(value: str) -> Optional["MudraCharacteristicUUID"]:
        for item in MudraCharacteristicUUID:
            if item.value == value:
                return item
        return None

class FirmwareCallbacks(Enum):
    STOP_ADVERTISING = 1
    FIRMWARE_CRASH = 2
    GESUTURE_RELEASE = 3
    FIRMWARE_STATUS = 4
    NAVIGATION_DELTA = 5
    NAVIGATION_DIRECTION = 6
    RESET_FUEL_GAUGE = 7
    PRESSURE = 8
    GESTURE_TAP = 9
    GESTURE_DOUBLE_TAP = 10
    GESTURE_TWIST = 11
    GESTURE_DOUBLE_TWIST = 12
    MAPPER_UPDATED = 13
    SAMPLE_TYPE_CHANGED = 14
    BAND_CONNECTION_STATUS = 15
    NONE = 16

    @property
    def description(self) -> str:
        return {
            FirmwareCallbacks.STOP_ADVERTISING: "StopAdvertising",
            FirmwareCallbacks.FIRMWARE_CRASH: "FirmwareCrash",
            FirmwareCallbacks.RESET_FUEL_GAUGE: "ResetFuelGauge",
            FirmwareCallbacks.GESUTURE_RELEASE: "GesutureRelease",
            FirmwareCallbacks.FIRMWARE_STATUS: "FirmwareStatus",
            FirmwareCallbacks.NAVIGATION_DELTA: "NavigationDelta",
            FirmwareCallbacks.NAVIGATION_DIRECTION: "NavigationDirection",
            FirmwareCallbacks.PRESSURE: "Pressure",
            FirmwareCallbacks.GESTURE_TAP: "GestureTap",
            FirmwareCallbacks.GESTURE_DOUBLE_TAP: "GestureDoubleTap",
            FirmwareCallbacks.GESTURE_TWIST: "GestureTwist",
            FirmwareCallbacks.GESTURE_DOUBLE_TWIST: "GestureDoubleTwist",
            FirmwareCallbacks.MAPPER_UPDATED: "mapper",
            FirmwareCallbacks.SAMPLE_TYPE_CHANGED: "sampleTypeChanged",
            FirmwareCallbacks.BAND_CONNECTION_STATUS: "bandConnectionStatus",
            FirmwareCallbacks.NONE: "none"
        }[self]

    # Static byte-array values as bytes objects
    REQUEST_FOR_MAPPER_BYTES = bytes([0x75, 0x03])
    STOP_ADVERTISING_BYTES = bytes([0x0a, 0x01])
    GESTURE_TAP_BYTES = bytes([0x61, 0x00])
    GESTURE_DOUBLE_TAP_BYTES = bytes([0x61, 0x01])
    GESTURE_TWIST_BYTES = bytes([0x61, 0x02])
    GESTURE_DOUBLE_TWIST_BYTES = bytes([0x61, 0x03])
    GESUTURE_RELEASE_BYTES = bytes([0x59, 0x01])
    RESET_FUEL_GAUGE_BYTES = bytes([0x93, 0x00])
    BAND_CONNECTION_STATUS_BYTES = bytes([0xbb, 0x01])

    MAPPER_UPDATED_LENGTH = 122

    @staticmethod
    def are_bytes_equal(bytes1: bytes, bytes2: bytes) -> bool:
        return bytes1 == bytes2

    @staticmethod
    def from_data(data: bytes) -> "FirmwareCallbacks":
        if len(data) >= 2 and FirmwareCallbacks.are_bytes_equal(data[:2], FirmwareCallbacks.STOP_ADVERTISING_BYTES.value):
            return FirmwareCallbacks.STOP_ADVERTISING
        if data and data[0] == 0x63:
            return FirmwareCallbacks.FIRMWARE_CRASH
        if data and data[0] == 0x03 and len(data) == FirmwareCallbacks.MAPPER_UPDATED_LENGTH:
            return FirmwareCallbacks.MAPPER_UPDATED
        if data and data[0] in (0x01, 0x02):
            return FirmwareCallbacks.FIRMWARE_STATUS
        if data and data[0] == 0x60:
            return FirmwareCallbacks.PRESSURE
        if len(data) >= 2 and FirmwareCallbacks.are_bytes_equal(data[:2], FirmwareCallbacks.GESTURE_TAP_BYTES.value):
            return FirmwareCallbacks.GESTURE_TAP
        if len(data) >= 2 and FirmwareCallbacks.are_bytes_equal(data[:2], FirmwareCallbacks.GESTURE_DOUBLE_TAP_BYTES.value):
            return FirmwareCallbacks.GESTURE_DOUBLE_TAP
        if len(data) >= 2 and FirmwareCallbacks.are_bytes_equal(data[:2], FirmwareCallbacks.GESTURE_TWIST_BYTES.value):
            return FirmwareCallbacks.GESTURE_TWIST
        if len(data) >= 2 and FirmwareCallbacks.are_bytes_equal(data[:2], FirmwareCallbacks.GESTURE_DOUBLE_TWIST_BYTES.value):
            return FirmwareCallbacks.GESTURE_DOUBLE_TWIST
        if len(data) >= 2 and FirmwareCallbacks.are_bytes_equal(data[:2], FirmwareCallbacks.GESUTURE_RELEASE_BYTES.value):
            return FirmwareCallbacks.GESUTURE_RELEASE
        if data and data[0] == 0x64:
            return FirmwareCallbacks.NAVIGATION_DELTA
        if data and data[0] == 0x22:
            return FirmwareCallbacks.SAMPLE_TYPE_CHANGED
        if data and data[0] == 0x65:
            return FirmwareCallbacks.NAVIGATION_DIRECTION
        if len(data) >= 2 and FirmwareCallbacks.are_bytes_equal(data[:2], FirmwareCallbacks.RESET_FUEL_GAUGE_BYTES.value):
            return FirmwareCallbacks.RESET_FUEL_GAUGE
        if len(data) >= 2 and FirmwareCallbacks.are_bytes_equal(data[:2], FirmwareCallbacks.BAND_CONNECTION_STATUS_BYTES.value):
            return FirmwareCallbacks.BAND_CONNECTION_STATUS
        return FirmwareCallbacks.NONE


class GestureGroup(Enum):
    GESTURE = "gesture"
    MOTION = "motion"

class GestureType(Enum):
    TAP = 0x00
    DOUBLE_TAP = 0x01
    TWIST = 0x02
    DOUBLE_TWIST = 0x03
    LEFT_PINCH = 0x04
    RIGHT_PINCH = 0x05
    UP_PINCH = 0x06
    DOWN_PINCH = 0x07
    REVERSE_LEFT_PINCH = 0x08
    REVERSE_RIGHT_PINCH = 0x09
    REVERSE_UP_PINCH = 0x0a
    REVERSE_DOWN_PINCH = 0x0b
    REVERSE_PINCH = 0x0c
    REVERSE_DOUBLE_TAP = 0x0d
    PINCH_ROLL_LEFT = 0x0e
    PINCH_ROLL_RIGHT = 0x0f
    PINCH_IN = 0x10
    PINCH_OUT = 0x11
    SHORT_TAP = 0x12
    REVERSE_TAP = 0x13

    @property
    def value_int(self):
        # To mirror Dart's .value int getter (just more explicit)
        return self.value

    @property
    def gesture_group(self):
        # Mirrors the gestureGroup getter in Dart
        if self in (
            GestureType.LEFT_PINCH,
            GestureType.RIGHT_PINCH,
            GestureType.UP_PINCH,
            GestureType.DOWN_PINCH,
            GestureType.REVERSE_LEFT_PINCH,
            GestureType.REVERSE_RIGHT_PINCH,
            GestureType.REVERSE_UP_PINCH,
            GestureType.REVERSE_DOWN_PINCH,
            GestureType.PINCH_ROLL_LEFT,
            GestureType.PINCH_ROLL_RIGHT,
            GestureType.PINCH_IN,
            GestureType.PINCH_OUT
        ):
            return GestureGroup.MOTION
        elif self in (
            GestureType.TAP,
            GestureType.DOUBLE_TAP,
            GestureType.TWIST,
            GestureType.DOUBLE_TWIST,
            GestureType.REVERSE_PINCH,
            GestureType.REVERSE_DOUBLE_TAP,
            GestureType.SHORT_TAP,
            GestureType.REVERSE_TAP,
        ):
            return GestureGroup.GESTURE
        else:
            return None

    @property
    def description(self):
        return {
            GestureType.TAP: "Tap",
            GestureType.DOUBLE_TAP: "Double Tap",
            GestureType.TWIST: "Twist",
            GestureType.DOUBLE_TWIST: "Double Twist",
            GestureType.LEFT_PINCH: "Left Pinch",
            GestureType.RIGHT_PINCH: "Right Pinch",
            GestureType.UP_PINCH: "Up Pinch",
            GestureType.DOWN_PINCH: "Down Pinch",
            GestureType.REVERSE_LEFT_PINCH: "Reverse Left Pinch",
            GestureType.REVERSE_RIGHT_PINCH: "Reverse Right Pinch",
            GestureType.REVERSE_UP_PINCH: "Reverse Up Pinch",
            GestureType.REVERSE_DOWN_PINCH: "Reverse Down Pinch",
            GestureType.REVERSE_PINCH: "Reverse Pinch",
            GestureType.REVERSE_DOUBLE_TAP: "Reverse Double Tap",
            GestureType.PINCH_ROLL_LEFT: "Pinch Roll Left",
            GestureType.PINCH_ROLL_RIGHT: "Pinch Roll Right",
            GestureType.PINCH_IN: "Pinch In",
            GestureType.PINCH_OUT: "Pinch Out",
            GestureType.SHORT_TAP: "Short Tap",
            GestureType.REVERSE_TAP: "Reverse Tap",
        }[self]

    @staticmethod
    def from_value(value: int) -> "GestureType":
        for type_ in GestureType:
            if type_.value == value:
                return type_
        raise ValueError(f"Invalid value for GestureType: {value:#04x}")

class ScrollDirection(Enum):
    normal = 0x00
    reverse = 0xFF

    @property
    def op_code(self) -> int:
        if self == ScrollDirection.normal:
            return 0x00
        elif self == ScrollDirection.reverse:
            return 0xFF

    @staticmethod
    def from_value(value: int) -> "ScrollDirection":
        if value == 0x00:
            return ScrollDirection.normal
        elif value == 0xFF:
            return ScrollDirection.reverse
        return ScrollDirection.reverse


class MapperMode(Enum):
    mouseMapper = 0x00
    keyboardMapper = 0x01

    @property
    def op_code(self) -> int:
        if self == MapperMode.mouseMapper:
            return 0x00
        elif self == MapperMode.keyboardMapper:
            return 0x01

    @staticmethod
    def from_value(value: int) -> "MapperMode":
        if value == 0x00:
            return MapperMode.mouseMapper
        elif value == 0x01:
            return MapperMode.keyboardMapper
        return MapperMode.mouseMapper


class BandMode(Enum):
    mudraBand = 0x00
    mudraLink = 0x01

    @property
    def op_code(self) -> int:
        if self == BandMode.mudraBand:
            return 0x00
        elif self == BandMode.mudraLink:
            return 0x01

    @staticmethod
    def from_value(value: int) -> "BandMode":
        if value == 0x00:
            return BandMode.mudraBand
        elif value == 0x01:
            return BandMode.mudraLink
        return None


class HandType(Enum):
    left = 0x00
    right = 0x01

    @property
    def op_code(self) -> int:
        if self == HandType.left:
            return 0x00
        elif self == HandType.right:
            return 0x01

class AirMouseButton(Enum):
    release = 0
    press = 1

    @property
    def description(self) -> str:
        if self == AirMouseButton.release:
            return "Release"
        elif self == AirMouseButton.press:
            return "Press"

    @staticmethod
    def from_value(value: int) -> "AirMouseButton":
        if value == 0:
            return AirMouseButton.release
        return AirMouseButton.press



class FirmwareDataType(Enum):
    snc = 0
    imuQuaternion = 2
    imuGyro = 3
    imuAccelometer = 4
    navigation = 5
    embeddedGesture = 6
    embeddedPressure = 7
    embeddedAirTouch = 8

    @property
    def value_int(self) -> int:
        return self.value

    @staticmethod
    def from_value(value: int) -> "FirmwareDataType":
        for item in FirmwareDataType:
            if item.value == value:
                return item
        return None

class ModelType(Enum):
    basic = 0
    basicWithoutQuaternion = 1
    neuralClicker = 2
    embedded = 3
    weightEstimation = 4

    @property
    def value_int(self) -> int:
        return self.value

    @staticmethod
    def from_value(value: int) -> "ModelType":
        for item in ModelType:
            if item.value == value:
                return item
        return None


class FirmwareCommand(Enum):
    enableSnc = 0
    disableSnc = 1
    getSerialNumber = 2
    enableImuNorm = 3
    disableImuNorm = 4
    enableImuQuaternion = 5
    disableImuQuaternion = 6
    enableImuGyro = 7
    disableImuGyro = 8
    enableImuAccRaw = 9
    disableImuAccRaw = 10
    getFirmwareVersion = 11
    getSecondaryDevicesState = 12
    pairSecondaryDevice = 13
    unpairSecondaryDevice = 14
    changeSecondaryDevice = 15
    enableGesture = 16
    disableGesture = 17
    enablePressure = 18
    disablePressure = 19
    enableNavigation = 20
    disableNavigation = 21
    enableEmbeddedAirTouch = 22
    disableEmbeddedAirTouch = 23
    getGeneralStatus = 24
    getAirMouseStatus = 25
    setSampleType = 26
    resetFuelGauge = 27
    sendFirmwareOutputTo = 28
    setDeviceMode = 29
    setBandMode = 30
    setMapperMode = 31
    resetMapperMode = 32
    airMouseSpeed = 33
    resetAirMouse = 34
    objectScale = 35
    recenterImuCubic = 36
    stopAdvertising = 37
    buttonsCommand = 38
    keyBoardCommand = 39
    handType = 40
    genericCommand = 41
    shippingMode = 42
    batteryInformation = 43
    updateHidMapping = 44
    getMouseHidMapper = 45
    getKeyboardHidMapper = 46
    enterToDfuMode = 47
    scrollSpeed = 48
    hidFrequency = 49
    scrollDirection = 50
    getScrollDirection = 51
    getMapperMode = 52

    @property
    def description(self) -> str:
        return {
            FirmwareCommand.enableSnc: "EnableSnc",
            FirmwareCommand.disableSnc: "DisableSnc",
            FirmwareCommand.getSerialNumber: "GetSerialNumber",
            FirmwareCommand.enableImuNorm: "EnableImuNorm",
            FirmwareCommand.disableImuNorm: "DisableImuNorm",
            FirmwareCommand.enableImuQuaternion: "EnableImuQuaternion",
            FirmwareCommand.disableImuQuaternion: "DisableImuQuaternion",
            FirmwareCommand.enableImuGyro: "EnableImuGyro",
            FirmwareCommand.disableImuGyro: "DisableImuGyro",
            FirmwareCommand.enableImuAccRaw: "EnableImuAccRaw",
            FirmwareCommand.disableImuAccRaw: "DisableImuAccRaw",
            FirmwareCommand.getFirmwareVersion: "GetFirmwareVersion",
            FirmwareCommand.getSecondaryDevicesState: "GetSecondaryDevicesState",
            FirmwareCommand.pairSecondaryDevice: "PairSecondaryDevice",
            FirmwareCommand.unpairSecondaryDevice: "UnpairSecondaryDevice",
            FirmwareCommand.changeSecondaryDevice: "ChangeSecondaryDevice",
            FirmwareCommand.enableGesture: "EnableGesture",
            FirmwareCommand.disableGesture: "DisableGesture",
            FirmwareCommand.enablePressure: "EnablePressure",
            FirmwareCommand.disablePressure: "DisablePressure",
            FirmwareCommand.enableNavigation: "EnableNavigation",
            FirmwareCommand.disableNavigation: "DisableNavigation",
            FirmwareCommand.enableEmbeddedAirTouch: "EnableEmbeddedAirTouch",
            FirmwareCommand.disableEmbeddedAirTouch: "DisableEmbeddedAirTouch",
            FirmwareCommand.getGeneralStatus: "GetGeneralStatus",
            FirmwareCommand.getAirMouseStatus: "GetAirMouseStatus",
            FirmwareCommand.setSampleType: "SetSampleType",
            FirmwareCommand.resetFuelGauge: "ResetFuelGauge",
            FirmwareCommand.sendFirmwareOutputTo: "SendFirmwareOutputTo",
            FirmwareCommand.setDeviceMode: "SetDeviceMode",
            FirmwareCommand.setBandMode: "SetBandMode",
            FirmwareCommand.setMapperMode: "SetMapperMode",
            FirmwareCommand.resetMapperMode: "ResetMapperMode",
            FirmwareCommand.airMouseSpeed: "AirMouseSpeed",
            FirmwareCommand.resetAirMouse: "ResetAirMouse",
            FirmwareCommand.objectScale: "ObjectScale",
            FirmwareCommand.recenterImuCubic: "RecenterImuCubic",
            FirmwareCommand.stopAdvertising: "StopAdvertising",
            FirmwareCommand.buttonsCommand: "ButtonsCommand",
            FirmwareCommand.keyBoardCommand: "KeyBoardCommand",
            FirmwareCommand.handType: "HandType",
            FirmwareCommand.genericCommand: "GenericCommand",
            FirmwareCommand.shippingMode: "ShippingMode",
            FirmwareCommand.batteryInformation: "BatteryInformation",
            FirmwareCommand.updateHidMapping: "UpdateHIDMapping",
            FirmwareCommand.getMouseHidMapper: "GetMouseHidMapper",
            FirmwareCommand.getKeyboardHidMapper: "GetKeyboardHidMapper",
            FirmwareCommand.enterToDfuMode: "EnterToDfuMode",
            FirmwareCommand.scrollSpeed: "ScrollSpeed",
            FirmwareCommand.hidFrequency: "HidFrequency",
            FirmwareCommand.scrollDirection: "ScrollDirection",
            FirmwareCommand.getScrollDirection: "GetScrollDirection",
            FirmwareCommand.getMapperMode: "GetMapperMode",
        }[self]

    @property
    def op_code(self) -> int:
        return self.value

    @property
    def id(self) -> bytes:
        from mudra_sdk.models.computation_wrapper import ComputationWrapper
        return ComputationWrapper.get_firmware_command_bytes(self.op_code)

    @staticmethod
    def from_value(value: int) -> Optional["FirmwareCommand"]:
        for item in FirmwareCommand:
            if item.value == value:
                return item
        return None

class FirmwareTarget(Enum):
    navigation_to_app = 0
    navigation_to_hid = 1
    gesture_to_hid = 2

    @property
    def op_code(self) -> int:
        return self.value

    @staticmethod
    def from_value(value: int) -> Optional["FirmwareTarget"]:
        for item in FirmwareTarget:
            if item.value == value:
                return item
        return None

class LicenseType(Enum):
    main = 0
    raw_data = 1
    tensorflow_data = 2
    double_tap = 3
    basic_model = 4

class HandType(Enum):
    left = 0
    right = 1

    @property
    def op_code(self) -> int:
        return self.value



"""
Platform-specific library loader for Mudra SDK
"""
import os
import sys
import platform
from pathlib import Path
from typing import Optional

try:
    import ctypes
    from ctypes import CDLL
    import ctypes.util
except ImportError:
    ctypes = None
    CDLL = None


def get_platform_info() -> tuple[str, str]:
    """
    Detect the current platform and architecture.
    
    Returns:
        tuple: (platform_name, architecture) where:
            - platform_name: 'windows', 'linux', 'darwin', or 'android'
            - architecture: 'x86', 'x64', 'x86_64', 'arm', 'arm64', 'aarch64'
    """
    system = platform.system().lower()
    machine = platform.machine().lower()
    
    # Detect Android
    if 'android' in system or hasattr(sys, 'getandroidapilevel'):
        platform_name = 'android'
    elif system == 'windows':
        platform_name = 'windows'
    elif system == 'darwin':
        platform_name = 'darwin'
    elif system == 'linux':
        platform_name = 'linux'
    else:
        platform_name = system
    
    # Normalize architecture names
    if machine in ('x86_64', 'amd64'):
        if platform_name == 'windows':
            arch = 'x64'
        elif platform_name == 'android':
            arch = 'x86_64'  # Android uses x86_64
        else:
            arch = 'x86_64'
    elif machine in ('i386', 'i686', 'x86'):
        arch = 'x86'
    elif machine in ('arm64', 'aarch64'):
        if platform_name == 'darwin':
            arch = 'arm64'
        elif platform_name == 'android':
            arch = 'aarch64'  # Android typically uses aarch64
        else:
            arch = 'aarch64'
    elif machine.startswith('arm'):
        if platform_name == 'android':
            arch = 'arm'  # Android armeabi-v7a maps to arm
        else:
            arch = 'arm'
    else:
        arch = machine
    
    return platform_name, arch


def get_library_extension(platform_name: str) -> str:
    """
    Get the library file extension for the given platform.
    
    Args:
        platform_name: Platform name ('windows', 'linux', 'darwin', 'android')
    
    Returns:
        Library file extension ('.dll', '.so', '.dylib')
    """
    extensions = {
        'windows': '.dll',
        'linux': '.so',
        'darwin': '.dylib',
        'android': '.so',
    }
    return extensions.get(platform_name, '.so')


def find_library_path(library_name: str, base_name: Optional[str] = None) -> Optional[Path]:
    """
    Find the path to the native library file.
    
    Args:
        library_name: Name of the library (e.g., 'MudraSDK')
        base_name: Optional base name if different from library_name
    
    Returns:
        Path to the library file, or None if not found
    """
    if base_name is None:
        base_name = library_name
    
    # Get platform info
    platform_name, arch = get_platform_info()
    ext = get_library_extension(platform_name)
    
    # Get the package directory
    package_dir = Path(__file__).parent.parent
    libs_dir = package_dir / 'libs'
    
    # Map platform names to directory names
    platform_dirs = {
        'windows': 'windows',
        'linux': 'linux',
        'darwin': 'darwin',
        'android': 'android',  # You may want to add android directory
    }
    
    platform_dir = platform_dirs.get(platform_name)
    if not platform_dir:
        return None
    
    # Try to find the library
    lib_path = libs_dir / platform_dir / arch / f"{base_name}{ext}"
    
    if lib_path.exists():
        return lib_path
    
    # Fallback: try without architecture subdirectory
    lib_path = libs_dir / platform_dir / f"{base_name}{ext}"
    if lib_path.exists():
        return lib_path
    
    return None


def load_library(library_name: str, base_name: Optional[str] = None) -> Optional[CDLL]:
    """
    Load the native library for the current platform.
    
    Args:
        library_name: Name of the library (e.g., 'MudraSDK')
        base_name: Optional base name if different from library_name
    
    Returns:
        CDLL object if successful, None otherwise
    
    Raises:
        OSError: If the library cannot be loaded
    """
    if ctypes is None or CDLL is None:
        raise ImportError("ctypes is not available on this platform")
    
    lib_path = find_library_path(library_name, base_name)
    
    if lib_path is None:
        platform_name, arch = get_platform_info()
        raise FileNotFoundError(
            f"Library '{library_name}' not found for platform '{platform_name}' "
            f"architecture '{arch}'. Expected path: {lib_path}"
        )
    
    try:
        # Load the library
        lib = CDLL(str(lib_path))
        return lib
    except OSError as e:
        raise OSError(
            f"Failed to load library '{library_name}' from {lib_path}: {e}"
        ) from e


# Convenience function for common use case
def get_mudra_library() -> Optional[CDLL]:
    """
    Load the Mudra core library.
    
    Returns:
        CDLL object for the Mudra core library
    """
    return load_library('MudraSDK', 'MudraSDK')




from abc import ABC, abstractmethod

from bleak.backends.device import BLEDevice

from mudra_sdk.models.enums import AirMouseButton, GestureType
from ..service.ble_service import MudraCharacteristicUUID
import mudra_sdk.models.mudra_device as mudraDeviceModule

from typing import Callable

OnBleDeviceDiscovered = Callable[[mudraDeviceModule.MudraDevice], None]  
OnBleDeviceConnected = Callable[[mudraDeviceModule.MudraDevice], None]
OnBleDeviceDisconnected = Callable[[mudraDeviceModule.MudraDevice], None]
OnBleDeviceDisconnecting = Callable[[mudraDeviceModule.MudraDevice], None]
OnBleDeviceConnecting = Callable[[mudraDeviceModule.MudraDevice], None]
OnBleDeviceConnectionFailed = Callable[[mudraDeviceModule.MudraDevice, str], None]
OnBluetoothStateChanged = Callable[[bool], None]

class MudraDelegate(ABC):
    @abstractmethod
    def on_device_discovered(self, device: mudraDeviceModule.MudraDevice):
        pass

    @abstractmethod
    def on_mudra_device_disconnected(self, device: mudraDeviceModule.MudraDevice):
        pass

    @abstractmethod
    def on_mudra_device_disconnecting(self, device: mudraDeviceModule.MudraDevice):
        pass

    @abstractmethod
    def on_mudra_device_connected(self, device: mudraDeviceModule.MudraDevice):
        pass

    @abstractmethod
    def on_mudra_device_connecting(self, device: mudraDeviceModule.MudraDevice):
        pass

    @abstractmethod
    def on_mudra_device_connection_failed(self, device: mudraDeviceModule.MudraDevice, error: str):
        pass

    @abstractmethod
    def on_bluetooth_state_changed(self, state: bool):
        pass



# --- Implementation of BleServiceDelegate abstract methods ---
class BleServiceDelegate(ABC):
    @abstractmethod
    def on_device_discovered(self, device: BLEDevice):
        pass

    @abstractmethod
    def on_mudra_device_disconnected(self, device: BLEDevice):
        pass

    @abstractmethod
    def on_mudra_device_disconnecting(self, device: BLEDevice):
        pass

    @abstractmethod
    def on_mudra_device_connected(self, device: BLEDevice):
        pass

    @abstractmethod
    def on_mudra_device_connecting(self, device: BLEDevice):
        pass

    @abstractmethod
    def on_mudra_device_connection_failed(self, device: BLEDevice, error: str):
        pass

    @abstractmethod
    def on_bluetooth_state_changed(self, state: bool):
        pass

    @abstractmethod
    def on_ble_characteristic_discovered(self, device: BLEDevice, characteristic_uuid: MudraCharacteristicUUID):
        pass

    @abstractmethod
    def on_pressure_data_received(self, device_address: str, pressure_data: int):
        pass
    
    @abstractmethod
    def on_gesture_data_received(self, device_address: str, gesture_type: GestureType):
        pass

    @abstractmethod
    def on_airmouse_button_changed_received(self, device_address: str, airmouse_button: AirMouseButton):
        pass

    @abstractmethod
    def on_firmware_status_updated(self, device_address: str, data: bytes):
        pass

    @abstractmethod
    def handle_snc(self, device_address: str, data: bytes):
        pass

    @abstractmethod
    def handle_imu(self, device_address: str, data: bytes):
        pass

    @abstractmethod
    def on_navigation_delta_received(self, device_address: str, delta_x: int, delta_y: int):
        pass

    @abstractmethod
    def on_charging_state_changed(self, device_address: str, is_charging: bool):
        pass

    @abstractmethod
    def on_serial_number_right_part_received(self, device_address: str, serial_number: int):
        pass

    @abstractmethod
    def on_serial_number_left_part_received(self, device_address: str, serial_number: int):
        pass

    @abstractmethod
    def on_firmware_version_received(self, device_address: str, version: str):
        pass

    @abstractmethod
    def on_battery_level_changed(self, device_address: str, level: int):
        pass


from bleak.backends.device import BLEDevice
from mudra_sdk.models.computation_wrapper import ComputationWrapper
from mudra_sdk.models.enums import AirMouseButton, GestureType, HandType, LicenseType, FirmwareTarget
from ..service import BleService, MudraCharacteristicUUID
from ..libs import load_library
from ctypes import CDLL
from typing import Optional

import mudra_sdk.models.mudra_device as mudraDeviceModule
import requests

from ..models.callbacks import BleServiceDelegate, MudraDelegate

class Mudra(BleServiceDelegate):
    _instance = None
    _delegate: Optional[MudraDelegate] = None
    _ble_service: Optional[BleService] = None

    _mudra_devices: dict[str, mudraDeviceModule.MudraDevice] = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            # Mark instance as not yet initialized; used to guard __init__
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        # Guard against running initialization logic multiple times on the singleton
        if getattr(self, "_initialized", False):
            return

        self._native_lib: Optional[CDLL] = None
        self._load_native_library()
        self._ble_service = BleService(delegate=self)
        self._initialized = True

       
    def get_license_for_email_from_cloud(self, email):
        url = ComputationWrapper.get_licenses_request_url()
        payload = {"email": email}
        try:
            response = requests.post(url, json=payload)
            # Parse the JSON response
            resp_json = response.json()
            # Iterate through all possible License items
            for license_item in LicenseType:
                key = ComputationWrapper.get_licenses_key(license_item.value)
                # Check if the key exists in response and its value is True
                if key in resp_json and resp_json[key] is True:
                    ComputationWrapper.set_license(license_item.value, key)
                    print(f"Set license: {license_item.name} with key: {key}")
                else:
                    print(f"License: {license_item.name}, Key: {key} not present or not True")
            return response
        except Exception as e:
            print("Request failed:", e)
            return None


    def _load_native_library(self):
        """
        Load the native library for the current platform.
        This will automatically detect the platform and load the correct .dll/.so/.dylib
        """
        try:
            self._native_lib = load_library('MudraSDK', 'MudraSDK')
            print(f"Native library loaded: {self._native_lib}")
        except (FileNotFoundError, OSError) as e:
            # Handle the case where the library is not found or cannot be loaded
            print(f"Warning: Could not load native library: {e}")
            self._native_lib = None

    @property
    def native_lib(self) -> Optional[CDLL]:
        """Get the loaded native library, or None if not loaded."""
        return self._native_lib

    def get_native_library(self) -> Optional[CDLL]:
        return self._native_lib

    async def send_general_command(self, device: mudraDeviceModule.MudraDevice, command: bytes):
        await self._ble_service.send_general_command(device, command)

    def get_firmware_command(self, command: int) -> bytes:
        return ComputationWrapper.get_firmware_command_bytes(command)

    ### ----------------------- Connection Methods ----------------------- ###

    async def connect(self, device: mudraDeviceModule.MudraDevice):
        await self._ble_service.connect(device)

    async def disconnect(self, device: mudraDeviceModule.MudraDevice):
        await self._ble_service.disconnect(device)

    ### ----------------------- Scan Methods ----------------------- ###

    async def scan(self):
        await self._ble_service.scan()

    async def stop_scan(self):
        await self._ble_service.stop_scanning()
        
    ### ----------------------- Setter Methods ----------------------- ###
    async def set_firmware_target(self, device: mudraDeviceModule.MudraDevice, firmware_target: FirmwareTarget, active: bool):
        await self._ble_service.set_firmware_target(device, firmware_target, active)

    async def set_hand(self, device: mudraDeviceModule.MudraDevice, hand_type: HandType):
        await self._ble_service.set_hand(device, hand_type)
        
    # --- Implementation of BleServiceDelegate abstract methods ---
    def on_ble_characteristic_discovered(self, device: mudraDeviceModule.MudraDevice, characteristic_uuid: MudraCharacteristicUUID):
        if device.address in self._mudra_devices:
            print(f"on_ble_characteristic_discovered: {device.name}")
            self._mudra_devices[device.address].on_characteristic_discovered(characteristic_uuid)
        else:
            print(f"on_ble_characteristic_discovered: Device not found: {device.name}")

    def on_pressure_data_received(self, device_address: str, pressure_data: int):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].on_pressure_data_received(pressure_data)

    def on_gesture_data_received(self, device_address: str, gesture_type: GestureType):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].on_gesture_data_received(gesture_type)

    def on_airmouse_button_changed_received(self, device_address: str, airmouse_button: AirMouseButton):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].on_airmouse_button_changed_received(airmouse_button)

    def on_firmware_status_updated(self, device_address: str, data: bytes):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].on_firmware_status_updated(data)

    def handle_snc(self, device_address: str, data: bytes):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].handle_snc(data)

    def handle_imu(self, device_address: str, data: bytes):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].handle_imu(data)

    def on_navigation_delta_received(self, device_address: str, delta_x: int, delta_y: int):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].on_navigation_delta_received(delta_x, delta_y)

    def on_charging_state_changed(self, device_address: str, is_charging: bool):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].on_charging_state_changed(is_charging)

    def on_serial_number_right_part_received(self, device_address: str, serial_number: int):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].on_serial_number_right_part_received(serial_number)

    def on_serial_number_left_part_received(self, device_address: str, serial_number: int):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].on_serial_number_left_part_received(serial_number)

    def on_firmware_version_received(self, device_address: str, version: str):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].on_firmware_version_received(version)

    def on_battery_level_changed(self, device_address: str, level: int):
        if device_address in self._mudra_devices:
            self._mudra_devices[device_address].on_battery_level_changed(level)
    # --- Implementation of MudraDelegate abstract methods ---
    
    def _create_device(self, device: BLEDevice) -> mudraDeviceModule.MudraDevice:
        if device.address in self._mudra_devices:
            return self._mudra_devices[device.address]
        mudra_device = mudraDeviceModule.MudraDevice(device)
        self._mudra_devices[device.address] = mudra_device
        return mudra_device

    def set_delegate(self, delegate: MudraDelegate):
        self._delegate = delegate

    def on_device_discovered(self, device: BLEDevice):
        mudra_device = self._create_device(device)
        if self._delegate:
            self._delegate.on_device_discovered(mudra_device)

    def on_mudra_device_disconnected(self, device: BLEDevice):
        mudra_device = self._mudra_devices[device.address]
        if mudra_device and self._delegate:
            self._delegate.on_mudra_device_disconnected(device)

    def on_mudra_device_disconnecting(self, device: BLEDevice):
        mudra_device = self._mudra_devices[device.address]
        if mudra_device and self._delegate:
            self._delegate.on_mudra_device_disconnecting(device)

    def on_mudra_device_connected(self, device: BLEDevice):
        mudra_device = self._mudra_devices[device.address]
        if mudra_device and self._delegate:
            self._delegate.on_mudra_device_connected(device)

    def on_mudra_device_connecting(self, device: BLEDevice):
        mudra_device = self._mudra_devices[device.address]
        if mudra_device and self._delegate:
            self._delegate.on_mudra_device_connecting(device)

    def on_mudra_device_connection_failed(self, device: BLEDevice, error: str):
        mudra_device = self._mudra_devices[device.address]
        if mudra_device and self._delegate:
            self._delegate.on_mudra_device_connection_failed(device, error)

    def on_bluetooth_state_changed(self, state: bool):
        if self._delegate:
            self._delegate.on_bluetooth_state_changed(state)




from typing import Optional, List, Any
from .enums import MapperMode, HandType, BandMode, ScrollDirection

# Dummy logger as Python equivalent for AppLoggers.firmwareStatus
class _DummyLogger:
    def debug(self, msg: str):
        print(msg)

class FirmwareStatus:
    _logger = _DummyLogger()

    # Status byte arrays
    general_status: Optional[List[int]]
    air_mouse_status: Optional[List[int]]

    # Firmware general status indexes
    SNC_ENABLED_INDEX = 14
    ACC_ENABLED_INDEX = 6
    GYRO_ENABLED_INDEX = 5
    QUATERNION_ENABLED_INDEX = 4
    NAVIGATION_ENABLED_INDEX = 10
    GESTURE_ENABLED_INDEX = 11
    PRESSURE_ENABLED_INDEX = 15
    AIR_TOUCH_ENABLED_INDEX = 19

    # Air mouse status indexes
    DEVICE_MODE_INDEX = 1
    SENDS_NAVIGATION_TO_HID_ENABLED_INDEX = 2
    SENDS_NAVIGATION_TO_APP_ENABLED_INDEX = 3
    SENDS_GESTURE_TO_HID_ENABLED_INDEX = 4
    CURRENT_MAPPER_MODE_INDEX = 5
    HAND_INDEX = 9
    POINTER_SPEED_X_INDEX = 10
    POINTER_SPEED_Y_INDEX = 11
    HID_FREQUENCY_INDEX = 14
    SCROLL_SPEED_INDEX = 15
    SCROLL_DIRECTION_INDEX = 16
    BAND_MODE_INDEX = 17

    def __init__(self):
        self.general_status = None
        self.air_mouse_status = None

    def update(self, data: bytes):
        if not data:
            return
        d = list(data)
        if d[0] == 1:
            old_status = self.general_status
            self.general_status = d

            if old_status is not None:
                if old_status[self.NAVIGATION_ENABLED_INDEX] != d[self.NAVIGATION_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : Navigation enabled changed: {d[self.NAVIGATION_ENABLED_INDEX] == 1}")
                if old_status[self.GESTURE_ENABLED_INDEX] != d[self.GESTURE_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : Gesture enabled changed: {d[self.GESTURE_ENABLED_INDEX] == 1}")
                if old_status[self.PRESSURE_ENABLED_INDEX] != d[self.PRESSURE_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : Pressure enabled changed: {d[self.PRESSURE_ENABLED_INDEX] == 1}")
                if old_status[self.SNC_ENABLED_INDEX] != d[self.SNC_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : SNC enabled changed: {d[self.SNC_ENABLED_INDEX] == 1}")
                if old_status[self.ACC_ENABLED_INDEX] != d[self.ACC_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : Accelerometer enabled changed: {d[self.ACC_ENABLED_INDEX] == 1}")
                if old_status[self.GYRO_ENABLED_INDEX] != d[self.GYRO_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : Gyro enabled changed: {d[self.GYRO_ENABLED_INDEX] == 1}")
                if old_status[self.QUATERNION_ENABLED_INDEX] != d[self.QUATERNION_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : Quaternion enabled changed: {d[self.QUATERNION_ENABLED_INDEX] == 1}")
                if old_status[self.AIR_TOUCH_ENABLED_INDEX] != d[self.AIR_TOUCH_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : AirTouch enabled changed: {d[self.AIR_TOUCH_ENABLED_INDEX] == 1}")
        elif d[0] == 2:
            old_status = self.air_mouse_status
            self.air_mouse_status = d

            if old_status is not None:
                if old_status[self.DEVICE_MODE_INDEX] != d[self.DEVICE_MODE_INDEX]:
                    self._logger.debug(f"Firmware status : Device mode changed: {d[self.DEVICE_MODE_INDEX]}")
                if old_status[self.SENDS_NAVIGATION_TO_APP_ENABLED_INDEX] != d[self.SENDS_NAVIGATION_TO_APP_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : Sends navigation to app changed: {d[self.SENDS_NAVIGATION_TO_APP_ENABLED_INDEX] == 1}")
                if old_status[self.SENDS_NAVIGATION_TO_HID_ENABLED_INDEX] != d[self.SENDS_NAVIGATION_TO_HID_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : Sends navigation to HID changed: {d[self.SENDS_NAVIGATION_TO_HID_ENABLED_INDEX] == 1}")
                if old_status[self.SENDS_GESTURE_TO_HID_ENABLED_INDEX] != d[self.SENDS_GESTURE_TO_HID_ENABLED_INDEX]:
                    self._logger.debug(f"Firmware status : Sends gesture to HID changed: {d[self.SENDS_GESTURE_TO_HID_ENABLED_INDEX] == 1}")
                if old_status[self.CURRENT_MAPPER_MODE_INDEX] != d[self.CURRENT_MAPPER_MODE_INDEX]:
                    self._logger.debug(f"Firmware status : mapper changed: {MapperMode.from_value(d[self.CURRENT_MAPPER_MODE_INDEX])}")
                if old_status[self.HAND_INDEX] != d[self.HAND_INDEX]:
                    self._logger.debug(f"Firmware status : hand changed: {'left' if d[self.HAND_INDEX] == 1 else 'right'}")
                if old_status[self.POINTER_SPEED_X_INDEX] != d[self.POINTER_SPEED_X_INDEX]:
                    self._logger.debug(f"Firmware status : pointer speed x changed: {d[self.POINTER_SPEED_X_INDEX]}")
                if old_status[self.POINTER_SPEED_Y_INDEX] != d[self.POINTER_SPEED_Y_INDEX]:
                    self._logger.debug(f"Firmware status : pointer speed y changed: {d[self.POINTER_SPEED_Y_INDEX]}")
                if old_status[self.SCROLL_SPEED_INDEX] != d[self.SCROLL_SPEED_INDEX]:
                    self._logger.debug(f"Firmware status : scrollSpeed changed: {d[self.SCROLL_SPEED_INDEX]}")
                if old_status[self.HID_FREQUENCY_INDEX] != d[self.HID_FREQUENCY_INDEX]:
                    self._logger.debug(f"Firmware status : hidFrequency changed: {d[self.HID_FREQUENCY_INDEX]}")
                if old_status[self.SCROLL_DIRECTION_INDEX] != d[self.SCROLL_DIRECTION_INDEX]:
                    self._logger.debug(f"Firmware status : scrollDirection changed: {d[self.SCROLL_DIRECTION_INDEX]}")
                if old_status[self.BAND_MODE_INDEX] != d[self.BAND_MODE_INDEX]:
                    self._logger.debug(f"Firmware status : bandMode changed: {d[self.BAND_MODE_INDEX]}")

    @property
    def is_navigation_enabled(self) -> bool:
        return bool(self.general_status and self.general_status[self.NAVIGATION_ENABLED_INDEX] == 1)

    @property
    def device_mode(self) -> int:
        if self.air_mouse_status:
            return self.air_mouse_status[self.DEVICE_MODE_INDEX]
        return -1

    @property
    def is_gesture_enabled(self) -> bool:
        return bool(self.general_status and self.general_status[self.GESTURE_ENABLED_INDEX] == 1)

    @property
    def is_pressure_enabled(self) -> bool:
        return bool(self.general_status and self.general_status[self.PRESSURE_ENABLED_INDEX] == 1)

    @property
    def is_snc_enabled(self) -> bool:
        return bool(self.general_status and self.general_status[self.SNC_ENABLED_INDEX] == 1)

    @property
    def is_acc_enabled(self) -> bool:
        return bool(self.general_status and self.general_status[self.ACC_ENABLED_INDEX] == 1)

    @property
    def is_gyro_enabled(self) -> bool:
        return bool(self.general_status and self.general_status[self.GYRO_ENABLED_INDEX] == 1)

    @property
    def is_quaternion_enabled(self) -> bool:
        return bool(self.general_status and self.general_status[self.QUATERNION_ENABLED_INDEX] == 1)

    @property
    def is_sends_navigation_to_app_enabled(self) -> bool:
        return bool(self.air_mouse_status and self.air_mouse_status[self.SENDS_NAVIGATION_TO_APP_ENABLED_INDEX] == 1)

    @property
    def is_sends_navigation_to_hid_enabled(self) -> bool:
        return bool(self.air_mouse_status and self.air_mouse_status[self.SENDS_NAVIGATION_TO_HID_ENABLED_INDEX] == 1)

    @property
    def is_sends_gesture_to_hid_enabled(self) -> bool:
        return bool(self.air_mouse_status and self.air_mouse_status[self.SENDS_GESTURE_TO_HID_ENABLED_INDEX] == 1)

    @property
    def is_air_touch_enabled(self) -> bool:
        return bool(self.general_status and self.general_status[self.AIR_TOUCH_ENABLED_INDEX] == 1)

    @property
    def hand_type(self) -> HandType:
        if self.air_mouse_status and self.air_mouse_status[self.HAND_INDEX] == 1:
            return HandType.left
        else:
            return HandType.right

    @property
    def pointer_speed_x(self) -> int:
        if self.air_mouse_status:
            return self.air_mouse_status[self.POINTER_SPEED_X_INDEX]
        return 0

    @property
    def pointer_speed_y(self) -> int:
        if self.air_mouse_status:
            return self.air_mouse_status[self.POINTER_SPEED_Y_INDEX]
        return 0

    @property
    def scroll_speed(self) -> Optional[int]:
        if self.air_mouse_status:
            return self.air_mouse_status[self.SCROLL_SPEED_INDEX]
        return None

    @property
    def hid_frequency(self) -> Optional[int]:
        if self.air_mouse_status:
            return self.air_mouse_status[self.HID_FREQUENCY_INDEX]
        return None

    @property
    def band_mode(self) -> Optional[BandMode]:
        if self.air_mouse_status:
            return BandMode.from_value(self.air_mouse_status[self.BAND_MODE_INDEX])
        return None

    @property
    def current_mapper_mode(self) -> MapperMode:
        if self.air_mouse_status:
            return MapperMode.from_value(self.air_mouse_status[self.CURRENT_MAPPER_MODE_INDEX])
        return MapperMode.from_value(0)

    @property
    def scroll_direction(self) -> ScrollDirection:
        if self.air_mouse_status and self.air_mouse_status[self.SCROLL_DIRECTION_INDEX] == 1:
            return ScrollDirection.normal
        else:
            return ScrollDirection.reverse



import asyncio
from calendar import c
import cmd
from enum import Enum
from typing import Any, Callable, Optional, Dict
from bleak import BleakScanner, BleakClient
from bleak.backends.device import BLEDevice
from bleak.backends.service import BleakGATTServiceCollection
from mudra_sdk.models.enums import AirMouseButton, FirmwareCallbacks, FirmwareCommand, FirmwareTarget, GestureType, HandType, MudraBLEServicesUUID, MudraCharacteristicUUID

from ..models.callbacks import BleServiceDelegate

from typing import Optional

class BleService:
    _instance = None
    _delegate: Optional[BleServiceDelegate] = None

    charging_state_true = 0x7b 
    bar_code_digits = 1000000

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, delegate: BleServiceDelegate):
        if not hasattr(self, '_initialized'):
            self._discovered_devices: set[Any] = set()
            self._scan_task = None
            self._scanner: Optional[BleakScanner] = None
            self._delegate = delegate
            self._connected_clients: Dict[str, BleakClient] = {}  # address -> client
            self._initialized = True

    ### ----------------------- Connection Methods ----------------------- ###
    
    async def connect(self, device: BLEDevice):
        print(f"Connecting to device: {device.name}")
        address = device.address
        
        # Check if already connected
        if address in self._connected_clients:
            print(f"Device {address} is already connected")
            return
        
        try:
            # Notify connecting
            if self._delegate:
                self._delegate.on_mudra_device_connecting(device)
            
            # Create client with disconnect callback.
            client = BleakClient(device, disconnected_callback=lambda client: self._on_disconnect_callback(device))
            
            # Attempt connection
            print(f"Calling BleakClient.connect() for {address}")
            await client.connect()
            import asyncio
            await asyncio.sleep(10)
            
            # Store the connected client
            self._connected_clients[address] = client
            device.client = client 
            
            # Discover services
            print(f"Discovering services for {device.name}")
            services = client.services
            
            # Notify connected
            if self._delegate:
                self._delegate.on_mudra_device_connected(device)
            
            if not await self.init_ble_services(device, services):
                print(f"Failed to initialize BLE services for {device.name} ({address})")
                await self.disconnect(device)
                if self._delegate:
                    self._delegate.on_mudra_device_connection_failed(device, "Failed to initialize BLE services")
                return 
            
            # Optionally store services in the device object
            device.services = services

            print(f"Successfully connected to {device.name} ({address})")
            
        except Exception as e:
            error_msg = str(e)
            print(f"Failed to connect to {address}: {error_msg}")
            
            # Notify connection failed
            if self._delegate:
                self._delegate.on_mudra_device_connection_failed(device, error_msg)
            
            # Clean up if connection attempt failed
            if address in self._connected_clients:
                del self._connected_clients[address]

    async def init_ble_services(self, device: BLEDevice, services: BleakGATTServiceCollection):
        client = device.client  # Get the BleakClient from the device

        for service in services:
            service_enum = MudraBLEServicesUUID.from_value(service.uuid)
            if service_enum:
                print(f"  -> Recognized as: {service_enum.name}")

            for char in service.characteristics:
                char_enum = MudraCharacteristicUUID.from_value(char.uuid)
                print(f"    Properties: {char.properties}")

                # Check if characteristic supports notify or indicate
                can_notify = "notify" in char.properties
                can_indicate = "indicate" in char.properties

                if char_enum:
                    if self._delegate:
                        self._delegate.on_ble_characteristic_discovered(device, char_enum)

                try:
                    match char_enum:
                        case MudraCharacteristicUUID.SNC_CHARACTERISTIC:
                            if can_notify or can_indicate:
                                print(f"Subscribing to notifications for SNC characteristic: {char.uuid}")
                                await client.start_notify(char.uuid, lambda sender, data: self._snc_notification_handler(device, sender, data))

                        case MudraCharacteristicUUID.IMU_CHARACTERISTIC:
                            if can_notify or can_indicate:
                                print(f"Subscribing to notifications for IMU characteristic: {char.uuid}")
                                await client.start_notify(char.uuid, lambda sender, data: self._imu_notification_handler(device, sender, data))

                        case MudraCharacteristicUUID.COMMAND_CHARACTERISTIC:
                            if can_notify or can_indicate:
                                print(f"Subscribing to notifications for COMMAND characteristic: {char.uuid}")
                                await client.start_notify(char.uuid, lambda sender, data: self._command_notification_handler(device, sender, data))

                        case MudraCharacteristicUUID.LOGGING_CHARACTERISTIC:
                            if can_notify or can_indicate:
                                print(f"Subscribing to notifications for LOGGING characteristic: {char.uuid}")
                                await client.start_notify(char.uuid, lambda sender, data: self._logging_notification_handler(device, sender, data))

                        case MudraCharacteristicUUID.MESSAGE_CHARACTERISTIC:
                            if can_notify or can_indicate:
                                print(f"Subscribing to notifications for MESSAGE characteristic: {char.uuid}")
                                await client.start_notify(char.uuid, lambda sender, data: self._message_notification_handler(device, sender, data))

                        case MudraCharacteristicUUID.BATTERY_CHARACTERISTIC:
                            if can_notify or can_indicate:
                                print(f"Subscribing to notifications for BATTERY characteristic: {char.uuid}")
                                await client.start_notify(char.uuid, lambda sender, data: self._battery_notification_handler(device, sender, data))
                            print(f"Reading BATTERY characteristic: {char.uuid}")
                            try:
                                value = await client.read_gatt_char(char.uuid)
                                self._battery_notification_handler(device, char.uuid, value)
                            except Exception as e:
                                print(f"Failed to read BATTERY characteristic {char.uuid}: {e}")

                        case MudraCharacteristicUUID.FIRMWARE_CHARACTERISTIC:
                            print(f"Reading FIRMWARE characteristic: {char.uuid}")
                            try:
                                value = await client.read_gatt_char(char.uuid)
                                self._firmware_notification_handler(device, char.uuid, value)
                            except Exception as e:
                                print(f"Failed to read FIRMWARE characteristic {char.uuid}: {e}")

                        case MudraCharacteristicUUID.SERIAL_RIGHT_PART_CHARACTERISTIC:
                            print(f"Reading SERIAL RIGHT PART characteristic: {char.uuid}")
                            try:
                                value = await client.read_gatt_char(char.uuid)
                                self._serial_right_notification_handler(device, char.uuid, value)
                            except Exception as e:
                                print(f"Failed to read SERIAL RIGHT PART characteristic {char.uuid}: {e}")

                        case MudraCharacteristicUUID.SERIAL_LEFT_PART_CHARACTERISTIC:
                            print(f"Reading SERIAL LEFT PART characteristic: {char.uuid}")
                            try:
                                value = await client.read_gatt_char(char.uuid)
                                self._serial_left_notification_handler(device, char.uuid, value)
                            except Exception as e:
                                print(f"Failed to read SERIAL LEFT PART characteristic {char.uuid}: {e}")

                        case MudraCharacteristicUUID.DFU_WITHOUT_BONDS:
                            if can_indicate:  # DFU typically uses indications
                                print(f"Subscribing to indications for DFU WITHOUT BONDS characteristic: {char.uuid}")
                                await client.start_notify(char.uuid, lambda sender, data: self._dfu_notification_handler(device, sender, data))

                        case MudraCharacteristicUUID.CHARGING_STATE:
                            if can_notify or can_indicate:
                                print(f"Subscribing to notifications for CHARGING STATE characteristic: {char.uuid}")
                                await client.start_notify(char.uuid, lambda sender, data: self._charging_notification_handler(device, sender, data))
                            print(f"Reading CHARGING STATE characteristic: {char.uuid}")
                            try:
                                value = await client.read_gatt_char(char.uuid)
                                self._charging_notification_handler(device, char.uuid, value)
                            except Exception as e:
                                print(f"Failed to read CHARGING STATE characteristic {char.uuid}: {e}")
                        case _:
                            if char_enum:
                                print(f"No notification handler for: {char_enum.name}")
                            else:
                                print(f"Unrecognized characteristic: {char.uuid}")

                except Exception as e:
                    print(f"Failed to subscribe/read {char.uuid}: {e}")
                    return False
        return True

    # Notification handler examples
    def _snc_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle SNC characteristic notifications."""
        if self._delegate:
            self._delegate.handle_snc(device.address, data)

    def _imu_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle IMU characteristic notifications."""
        if self._delegate:
            self._delegate.handle_imu(device.address, data)

    def _command_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle COMMAND characteristic notifications."""
        address = device.address
        firmware_callback = FirmwareCallbacks.from_data(data)
        if firmware_callback != FirmwareCallbacks.BAND_CONNECTION_STATUS:
            print(f"COMMAND notification from {sender} (device {address}): {data.hex()}")
        match firmware_callback:
            case FirmwareCallbacks.STOP_ADVERTISING:
                print("STOP_ADVERTISING")
            case FirmwareCallbacks.FIRMWARE_CRASH:
                print("FIRMWARE_CRASH")
            case FirmwareCallbacks.RESET_FUEL_GAUGE:
                print("RESET_FUEL_GAUGE")
            case FirmwareCallbacks.GESUTURE_RELEASE:
                print("GESUTURE_RELEASE")
                if self._delegate:
                    self._delegate.on_airmouse_button_changed_received(device.address, AirMouseButton.release)
            case FirmwareCallbacks.FIRMWARE_STATUS:
                if self._delegate:
                    self._delegate.on_firmware_status_updated(device.address, data)
            case FirmwareCallbacks.NAVIGATION_DELTA:
                delta_x = self._convert_two_bytes_to_navigation_delta(data[2], data[1])
                delta_y = self._convert_two_bytes_to_navigation_delta(data[4], data[3])
                if self._delegate:
                    self._delegate.on_navigation_delta_received(device.address, delta_x, delta_y)
            case FirmwareCallbacks.NAVIGATION_DIRECTION:
                print("NAVIGATION_DIRECTION")
            case FirmwareCallbacks.PRESSURE:
                if self._delegate:
                    self._delegate.on_pressure_data_received(device.address, data[1])
            case FirmwareCallbacks.GESTURE_TAP:
                if self._delegate:
                    self._delegate.on_gesture_data_received(device.address, GestureType.TAP)
                    self._delegate.on_airmouse_button_changed_received(device.address, AirMouseButton.press)
            case FirmwareCallbacks.GESTURE_DOUBLE_TAP:
                if self._delegate:
                    self._delegate.on_gesture_data_received(device.address, GestureType.DOUBLE_TAP)
            case FirmwareCallbacks.GESTURE_TWIST:
                if self._delegate:
                    self._delegate.on_gesture_data_received(device.address, GestureType.TWIST)
            case FirmwareCallbacks.GESTURE_DOUBLE_TWIST:
                if self._delegate:
                    self._delegate.on_gesture_data_received(device.address, GestureType.DOUBLE_TWIST)
            case FirmwareCallbacks.MAPPER_UPDATED:
                print("MAPPER_UPDATED")
            case FirmwareCallbacks.SAMPLE_TYPE_CHANGED:
                print("SAMPLE_TYPE_CHANGED")
            case FirmwareCallbacks.BAND_CONNECTION_STATUS:
                pass
            case FirmwareCallbacks.NONE:
                print("NONE")
            case _:
                print("Unknown firmware callback")

    def _logging_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle LOGGING characteristic notifications."""
        print(f"LOGGING notification from {sender} (device {device.address}): {data.hex()}")

    def _message_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle MESSAGE characteristic notifications."""
        print(f"MESSAGE notification from {sender} (device {device.address}): {data.hex()}")

    def _battery_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle BATTERY characteristic notifications."""
        battery_level = int(data[0]) if len(data) > 0 else 0
        if self._delegate:
            self._delegate.on_battery_level_changed(device.address, battery_level)

    def _firmware_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle FIRMWARE characteristic notifications."""
        if self._delegate:
            self._delegate.on_firmware_version_received(device.address, data.decode('utf-8', errors='ignore'))

    def _serial_right_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle SERIAL RIGHT PART characteristic notifications."""
        try:
            string_serial_number = data.decode('utf-8', errors='ignore')
        except Exception:
            string_serial_number = ''.join([chr(b) for b in data])
        try:
            value = int(string_serial_number, 16)
        except Exception:
            value = 0

        if self._delegate:
            self._delegate.on_serial_number_right_part_received(device.address, value)

    def _serial_left_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle SERIAL LEFT PART characteristic notifications."""
        try:
            string_serial_number = data.decode('utf-8', errors='ignore')
        except Exception:
            string_serial_number = ''.join([chr(b) for b in data])
        try:
            value = int(string_serial_number, 16)
        except Exception:
            value = 0

        calculated_value = value * self.bar_code_digits if value != 0 else 0

        if self._delegate:
            self._delegate.on_serial_number_left_part_received(device.address, calculated_value)

    def _dfu_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle DFU WITHOUT BONDS characteristic notifications/indications."""
        print(f"DFU notification from {sender} (device {device.address}): {data.hex()}")

    def _charging_notification_handler(self, device: BLEDevice, sender: int, data: bytearray):
        """Handle CHARGING STATE characteristic notifications."""
        is_charging = data[0] == self.charging_state_true
        self._delegate.on_charging_state_changed(device.address, is_charging)


    def _convert_two_bytes_to_navigation_delta(self, byte1: int, byte2: int) -> int:
        """Convert two bytes (high, low) into a signed 16-bit navigation delta."""
        value = (byte1 << 8) | byte2
        if value & 0x8000:
            value = value - 0x10000  # Convert to negative for signed 16-bit
        return value
    
    async def disconnect(self, device: BLEDevice):
        address = device.address
        
        # Check if device is connected
        if address not in self._connected_clients:
            print(f"Device {address} is not connected")
            return
        
        try:
            # Notify disconnecting
            if self._delegate:
                self._delegate.on_mudra_device_disconnecting(device)
            
            client = self._connected_clients[address]

            # If it's already disconnected (e.g. due to unexpected disconnect),
            # skip calling disconnect() to avoid spurious errors.
            if getattr(client, "is_connected", False):
                # Disconnect
                await client.disconnect()
            
            # Remove from connected clients
            del self._connected_clients[address]
            device.client = None  # Clear client reference if stored
            
            # Notify disconnected
            if self._delegate:
                self._delegate.on_mudra_device_disconnected(device)
            
            print(f"Successfully disconnected from {device.name} ({address})")
            
        except Exception as e:
            if address in self._connected_clients:
                del self._connected_clients[address]
            device.client = None

    def _on_disconnect_callback(self, device: BLEDevice):
        """Internal callback triggered when device disconnects unexpectedly."""
        address = device.address
        
        print(f"Device {address} disconnected unexpectedly")
        
        # Clean up
        if address in self._connected_clients:
            del self._connected_clients[address]
            device.client = None
        
        # Notify delegate
        if self._delegate:
            self._delegate.on_mudra_device_disconnected(device)

    def is_connected(self, device: BLEDevice) -> bool:
        """Check if a device is currently connected."""
        address = device.address
        if address not in self._connected_clients:
            return False
        
        client = self._connected_clients[address]
        return client.is_connected

    async def disconnect_all(self):
        """Disconnect all connected devices."""
        devices_to_disconnect = list(self._connected_clients.keys())
        
        for address in devices_to_disconnect:
            client = self._connected_clients[address]
            try:
                await client.disconnect()
            except Exception as e:
                print(f"Error disconnecting {address}: {e}")
            finally:
                if address in self._connected_clients:
                    del self._connected_clients[address]


    async def send_general_command(self, device: BLEDevice, command: bytes):
        try:
            client = device.client
            if client:
                print(f"Sending general command: {command.hex()}")
                await client.write_gatt_char(MudraCharacteristicUUID.COMMAND_CHARACTERISTIC.value, command)
            else:
                print(f"Device {device.address} is not connected")
        except Exception as e:
            print(f"Error sending generic command to {device.address}: {e}")


    ### ----------------------- Setter Methods ----------------------- ###

    async def set_firmware_target(self, device: BLEDevice, firmware_target: FirmwareTarget, active: bool):
        cmd = bytearray(FirmwareCommand.sendFirmwareOutputTo.id)
        cmd[1] = int(firmware_target.op_code)
        cmd[2] = int(bool(active))
        await self.send_general_command(device, bytes(cmd))

    async def set_hand(self, device: BLEDevice, hand_type: HandType):
        cmd = bytearray(FirmwareCommand.handType.id)
        cmd[1] = int(hand_type.op_code)
        await self.send_general_command(device, bytes(cmd))

    ### ----------------------- Scan Methods ----------------------- ###
    
    @property
    def _is_scanning(self) -> bool:
        """Check if currently scanning for devices."""
        return self._scan_task is not None and not self._scan_task.done()

    async def scan(self):
        """Start scanning for BLE devices."""
        if self._is_scanning:
            return
        print("Starting scan")
        self._discovered_devices.clear()
        self._scan_task = asyncio.create_task(self._scan_loop())

    async def stop_scanning(self):
        """Stop scanning for BLE devices."""
        if not self._is_scanning:
            return
        print("Stopping scan")
        
        # Cancel the scan task
        if self._scan_task and not self._scan_task.done():
            self._scan_task.cancel()
            try:
                await self._scan_task
            except asyncio.CancelledError:
                pass
        
        # Stop the scanner if it exists
        if self._scanner:
            try:
                await self._scanner.stop()
            except Exception:
                pass
            finally:
                self._scanner = None
        
        self._scan_task = None

    async def _scan_loop(self):
        """Internal method that performs the actual scanning."""
        try:
            self._scanner = BleakScanner(detection_callback=self._on_device_detected)
            
            await self._scanner.start()
            print("Scan started successfully")
            
            # Keep scanning until cancelled
            while True:
                await asyncio.sleep(1)
                
        except asyncio.CancelledError:
            print("Scan cancelled")
            if self._scanner:
                try:
                    await self._scanner.stop()
                except Exception:
                    pass
            raise
        except Exception as e:
            print(f"Error during scan: {e}")
            raise
        finally:
            self._scanner = None
            self._scan_task = None

    def _on_device_detected(self, device: BLEDevice, advertisement_data):
        """Internal callback when a BLE device is detected."""            
        if self._delegate:
            try:
                if device.name and "Mudra" in device.name:
                    # Use address as identifier, since address is unique per BLE device
                    if device.address not in self._discovered_devices:
                        self._discovered_devices.add(device.address)
                        self._delegate.on_device_discovered(device)
            except Exception as e:
                print(f"Error processing device {device.address}: {e}")


import asyncio
from typing import Callable, Optional
from bleak.backends.device import BLEDevice

from mudra_sdk.models.enums import AirMouseButton, FirmwareCommand, FirmwareDataType, FirmwareTarget, GestureType, HandType, ModelType
from mudra_sdk.models.firmware_status import FirmwareStatus
import mudra_sdk.models.mudra as mudraModule
import mudra_sdk.models.computation_wrapper as computationWrapperModule
from mudra_sdk.service.ble_service import MudraBLEServicesUUID, MudraCharacteristicUUID

class MudraDevice(BLEDevice):
    def __init__(self, ble_device: BLEDevice):
        super().__init__(ble_device.address, ble_device.name, ble_device.details)

        # Initialize characteristic status
        self._characteristic_status = {
            MudraCharacteristicUUID.SNC_CHARACTERISTIC: False,
            MudraCharacteristicUUID.IMU_CHARACTERISTIC: False,
            MudraCharacteristicUUID.COMMAND_CHARACTERISTIC: False,
            MudraCharacteristicUUID.LOGGING_CHARACTERISTIC: False,
            MudraCharacteristicUUID.MESSAGE_CHARACTERISTIC: False,
            MudraCharacteristicUUID.BATTERY_CHARACTERISTIC: False,
            MudraCharacteristicUUID.FIRMWARE_CHARACTERISTIC: False,
            MudraCharacteristicUUID.SERIAL_RIGHT_PART_CHARACTERISTIC: False,
            MudraCharacteristicUUID.SERIAL_LEFT_PART_CHARACTERISTIC: False,
            MudraCharacteristicUUID.DFU_WITHOUT_BONDS: False,
            MudraCharacteristicUUID.CHARGING_STATE: False,
        }

        # Initialize service status
        self._service_status = {
            MudraBLEServicesUUID.COMMAND_SERVICE: False,
            MudraBLEServicesUUID.BATTERY_SERVICE: False,
            MudraBLEServicesUUID.DFU_SERVICE: False,
            MudraBLEServicesUUID.INFORMATION_SERVICE: False,
        }

        self._state_enabled = {
            FirmwareDataType.snc: False,
            FirmwareDataType.imuQuaternion: False,
            FirmwareDataType.imuGyro: False,
            FirmwareDataType.imuAccelometer: False,
            FirmwareDataType.navigation: False,
            FirmwareDataType.embeddedGesture: False,
            FirmwareDataType.embeddedPressure: False,
            FirmwareDataType.embeddedAirTouch: False,
        }

        ### ----------------------- Ready Methods ----------------------- ###
        self._on_pressure_ready: Optional[computationWrapperModule.OnPressureCallback] = None
        self._on_gesture_ready: Optional[computationWrapperModule.OnGestureDataReceivedCallback] = None
        self._on_airmouse_button_changed_ready: Optional[computationWrapperModule.OnAirMouseButtonChangedReceivedCallback] = None
        self._on_navigation_ready: Optional[computationWrapperModule.OnNavigationDeltaReceivedCallback] = None

        ### ----------------------- Parameters Callbacks ----------------------- ###
        self._on_charging_state_changed: Optional[computationWrapperModule.OnChargingStateChangedCallback] = None
        self._on_battery_level_changed: Optional[computationWrapperModule.OnBatteryLevelChangedCallback] = None

        ### ----------------------- Computation Wrapper ----------------------- ###
        self.firmware_status = FirmwareStatus()
        self._computation_wrapper = computationWrapperModule.ComputationWrapper(0, "./models/model.tflite", "./models/model.tflite")
        self._computation_wrapper.set_model_type(ModelType.embedded)


    async def connect(self):
        await mudraModule.Mudra().connect(self)

    async def disconnect(self):
        await mudraModule.Mudra().disconnect(self)

    ### ----------------------- Ready Methods ----------------------- ###

    def on_pressure_data_received(self, pressure_data: int):
        if self._on_pressure_ready:
            self._on_pressure_ready(pressure_data)

    def on_gesture_data_received(self, gesture_type: GestureType):
        if self._on_gesture_ready:
            self._on_gesture_ready(gesture_type)

    def on_airmouse_button_changed_received(self, airmouse_button: AirMouseButton):
        if self._on_airmouse_button_changed_ready:
            self._on_airmouse_button_changed_ready(airmouse_button)

    def on_navigation_delta_received(self, delta_x: int, delta_y: int):
        if self._on_navigation_ready:
            self._on_navigation_ready(delta_x, delta_y)

    def on_charging_state_changed(self, is_charging: bool):
        self._is_charging = is_charging
        if self._on_charging_state_changed:
            self._on_charging_state_changed(is_charging)

    def on_serial_number_right_part_received(self, serial_number: int):
        self._serial_number_right_part = serial_number

    def on_serial_number_left_part_received(self, serial_number: int):
        self._serial_number_left_part = serial_number

    def on_firmware_version_received(self, version: str):
        self._firmware_version = version

    def on_battery_level_changed(self, level: int):
        self._battery_level = level
        if self._on_battery_level_changed:
            self._on_battery_level_changed(level)

    ### ----------------------- Getter Methods ----------------------- ###
    
    def get_is_charging(self) -> bool:
        return self._is_charging

    def get_serial_number(self) -> Optional[int]:
        if self._serial_number_right_part is not None and self._serial_number_left_part is not None:
            return self._serial_number_right_part + self._serial_number_left_part
        return None

    def get_firmware_version(self) -> Optional[str]:
        return self._firmware_version

    ### ----------------------- Setter Methods ----------------------- ###

    async def set_firmware_target(self, firmware_target: FirmwareTarget, active: bool):
        if self._characteristic_status[MudraCharacteristicUUID.COMMAND_CHARACTERISTIC]: 
            await mudraModule.Mudra().set_firmware_target(self, firmware_target, active)

    async def set_hand(self, hand_type: HandType):
        if self._characteristic_status[MudraCharacteristicUUID.COMMAND_CHARACTERISTIC]:
            await mudraModule.Mudra().set_hand(self, hand_type)

    async def set_on_charging_state_changed(self, callback: computationWrapperModule.OnChargingStateChangedCallback) -> None:
        self._on_charging_state_changed = callback

    async def set_on_battery_level_changed(self, callback: computationWrapperModule.OnBatteryLevelChangedCallback) -> None:
        self._on_battery_level_changed = callback

    ### ----------------------- Delegate Methods ----------------------- ###

    def on_characteristic_discovered(self, characteristic_uuid: MudraCharacteristicUUID):
        print(f"Characteristic discovered: {characteristic_uuid}")
        self._characteristic_status[characteristic_uuid] = True

    def on_firmware_status_updated(self, data: bytes):
        self.firmware_status.update(data)

    def _handle_data(self, data: bytes, error_message: str, native_handler_attr: str):
        try:
            self._computation_wrapper.handle_data(data, len(data))
        except Exception as e:
            print(f"Exception in _handle_data: {e}")

    def handle_snc(self, data: bytes):
        self._handle_data(data, "Failed to handle SNC. The handleSnc function is not available.", "handleSnc")

    def handle_imu(self, data: bytes):
        self._handle_data(data, "Failed to handle IMU. The handleImu function is not available.", "handleImu")

    async def set_on_snc_ready(self, callback: computationWrapperModule.OnRawDataCallback) -> None:
        self._on_snc_ready = callback
        self._computation_wrapper.set_on_snc_ready(callback)
        await self._update_state_enabled()
    
    async def set_on_imu_acc_ready(self, callback: computationWrapperModule.OnRawDataCallback) -> None:
        self._on_imu_acc_ready = callback
        self._computation_wrapper.set_on_imu_acc_ready(callback)
        await self._update_state_enabled()
    
    async def set_on_imu_gyro_ready(self, callback: computationWrapperModule.OnRawDataCallback) -> None:
        self._on_imu_gyro_ready = callback
        self._computation_wrapper.set_on_imu_gyro_ready(callback)
        await self._update_state_enabled()

    async def set_on_pressure_ready(self, callback: computationWrapperModule.OnPressureCallback) -> None:
        self._on_pressure_ready = callback
        self._computation_wrapper.set_on_pressure_ready(callback)
        await self._update_state_enabled()

    async def set_on_navigation_ready(self, callback: computationWrapperModule.OnNavigationDeltaReceivedCallback) -> None:
        self._on_navigation_ready = callback
        self._computation_wrapper.set_on_navigation_ready(callback != None)
        await self._update_state_enabled()

    async def set_on_gesture_ready(self, callback: computationWrapperModule.OnGestureDataReceivedCallback) -> None:
        self._on_gesture_ready = callback
        self._computation_wrapper.set_on_gesture_ready(callback)
        await self._update_state_enabled()

    async def set_on_button_changed(self, callback: computationWrapperModule.OnAirMouseButtonChangedReceivedCallback) -> None:
        self._on_airmouse_button_changed_ready = callback
        self._computation_wrapper.set_on_button_changed(callback)
        await self._update_state_enabled()

    async def set_air_touch_active(self, active: bool) -> None:
        self._on_embedded_airtouch_ready = active
        self._computation_wrapper.set_air_touch_active(active)
        await self._update_state_enabled()

    ### ----------------------- Is Callbacks set Methods ----------------------- ###

    def is_on_imu_gyro_callback_set(self) -> bool:
        return self._on_imu_gyro_ready is not None

    def is_on_snc_callback_set(self) -> bool:
        return self._on_snc_ready is not None

    def is_on_imu_acc_raw_callback_set(self) -> bool:
        return self._on_imu_acc_ready is not None

    def is_on_pressure_ready_set(self) -> bool:
        return self._on_pressure_ready is not None

    def is_on_gesture_callback_set(self) -> bool:
        return self._on_gesture_ready is not None

    def is_on_button_changed_callback_set(self) -> bool:
        return self._on_airmouse_button_changed_ready is not None

    def is_on_navigation_callback_set(self) -> bool:
        return self._on_navigation_ready is not None

    def is_on_charging_state_changed_callback_set(self) -> bool:
        return self._on_charging_state_changed is not None
        
    def is_on_battery_level_changed_callback_set(self) -> bool:
        return self._on_battery_level_changed is not None

    ### ----------------------- Is Data Needed ----------------------- ###
    
    async def _update_state_enabled(self):
        print("Updating state enabled for all firmware data types")
        for firmware_data_type in FirmwareDataType:
            await self._update_data_enabled(firmware_data_type)

    async def _update_data_enabled(self, firmware_data_type):

        is_needed = self._computation_wrapper.is_data_needed(firmware_data_type.value)
        print(f"Is data needed for {firmware_data_type.name} is {is_needed}")
        if is_needed != self._state_enabled.get(firmware_data_type):
            print(f"Updating state enabled for {firmware_data_type.name} to {is_needed}")
            self._state_enabled[firmware_data_type] = is_needed
            print(f"State enabled for {firmware_data_type.name} is {self._state_enabled.get(firmware_data_type)}")
            await self._update_configuration(firmware_data_type)


    async def get_firmware_command(self, command: int) -> bytes:
        return mudraModule.Mudra().get_firmware_command(command)

    async def _update_configuration(self, firmware_data_type):
        print(f"Updating configuration for {firmware_data_type.name}")
        if self._characteristic_status[MudraCharacteristicUUID.COMMAND_CHARACTERISTIC]:
            cmd: Optional[bytes] = None
            if firmware_data_type == FirmwareDataType.snc:
                print(f"Updating SNC configuration")
                cmd = FirmwareCommand.enableSnc.id if self._state_enabled.get(firmware_data_type) else FirmwareCommand.disableSnc.id
            elif firmware_data_type == FirmwareDataType.embeddedPressure:
                print(f"Updating Pressure configuration")
                cmd = FirmwareCommand.enablePressure.id if self._state_enabled.get(firmware_data_type) else FirmwareCommand.disablePressure.id
            elif firmware_data_type == FirmwareDataType.imuAccelometer:
                print(f"Updating IMU Acc configuration")
                cmd = FirmwareCommand.enableImuAccRaw.id if self._state_enabled.get(firmware_data_type) else FirmwareCommand.disableImuAccRaw.id
            elif firmware_data_type == FirmwareDataType.imuGyro:
                print(f"Updating IMU Gyro configuration")
                cmd = FirmwareCommand.enableImuGyro.id if self._state_enabled.get(firmware_data_type) else FirmwareCommand.disableImuGyro.id
            elif firmware_data_type == FirmwareDataType.imuQuaternion:
                print(f"Updating IMU Quaternion configuration")
                cmd = FirmwareCommand.enableImuQuaternion.id if self._state_enabled.get(firmware_data_type) else FirmwareCommand.disableImuQuaternion.id
            elif firmware_data_type == FirmwareDataType.navigation:
                print(f"Updating Navigation configuration")
                cmd = FirmwareCommand.enableNavigation.id if self._state_enabled.get(firmware_data_type) else FirmwareCommand.disableNavigation.id
            elif firmware_data_type == FirmwareDataType.embeddedGesture:
                print(f"Updating Embedded Gesture configuration")
                cmd = FirmwareCommand.enableGesture.id if self._state_enabled.get(firmware_data_type) else FirmwareCommand.disableGesture.id
            elif firmware_data_type == FirmwareDataType.embeddedPressure:
                print(f"Updating Embedded Pressure configuration")
                cmd = FirmwareCommand.enablePressure.id if self._state_enabled.get(firmware_data_type) else FirmwareCommand.disablePressure.id
            elif firmware_data_type == FirmwareDataType.embeddedAirTouch:
                print(f"Updating Embedded Air Touch configuration")
                cmd = FirmwareCommand.enableEmbeddedAirTouch.id if self._state_enabled.get(firmware_data_type) else FirmwareCommand.disableEmbeddedAirTouch.id
            else:
                print(f"Updating {firmware_data_type.name} configuration")
                
            if cmd is not None:
                await mudraModule.Mudra().send_general_command(self, cmd)
            


from ctypes import POINTER, c_float, c_int, c_void_p, c_char_p, c_uint8, c_int32, c_size_t, Structure, c_char
import ctypes
from typing import Callable, List, Optional
from mudra_sdk.models.enums import AirMouseButton, GestureType, ModelType
import mudra_sdk.models.mudra as mudraModule


OnGestureDataReceivedCallback = Callable[[GestureType], None]
OnAirMouseButtonChangedReceivedCallback = Callable[[AirMouseButton], None]
OnNavigationDeltaReceivedCallback = Callable[[int, int], None]

OnChargingStateChangedCallback = Callable[[bool], None]
OnBatteryLevelChangedCallback = Callable[[int], None]

ComputationHandle = c_void_p
OnRawDataCallback = Callable[[int, List[float], int, float, List[float]], None]
OnRawDataCallbackC = ctypes.CFUNCTYPE(None, c_int, POINTER(c_float), c_int, c_int, c_float, POINTER(c_float))

OnPressureCallback = Callable[[int], None]
OnPressureCallbackC = ctypes.CFUNCTYPE(None, c_float)

GestureCallbackC = ctypes.CFUNCTYPE(None, c_int)
OnAirMouseButtonChangedCallbackC = ctypes.CFUNCTYPE(None, c_uint8)


class ByteBuffer(Structure):
    _fields_ = [
        ("data", POINTER(c_uint8)),
        ("length", c_size_t),
    ]


class ComputationWrapper:
    _native_lib = None
    _snc_rms_count = 3
    
    def __init__(self, index: int, model_path: str = "./", writable_path: str = "./"):
        self._ensure_native_lib_loaded()
        self._initialize_native_lib()

        self._handle = ComputationWrapper._native_lib.create_computation(
            index,
            model_path.encode(),
            writable_path.encode(),
        )
        if not self._handle:
            raise RuntimeError("Failed to create computation")

    @classmethod
    def _ensure_native_lib_loaded(cls):
        if cls._native_lib is None:
            cls._native_lib = mudraModule.Mudra().get_native_library()
        if cls._native_lib is None:
            raise RuntimeError("Native library is not loaded")

    @classmethod
    def _initialize_native_lib(cls):
        lib = cls._native_lib
        
        lib.create_computation.argtypes = [c_int, c_char_p, c_char_p]
        lib.create_computation.restype = ComputationHandle

        lib.handle_data.argtypes = [ComputationHandle, POINTER(c_uint8), c_int32]
        lib.handle_data.restype = None

        lib.is_data_needed.argtypes = [ComputationHandle, c_int]
        lib.is_data_needed.restype = ctypes.c_bool

        lib.set_on_snc_ready.argtypes = [ComputationHandle, OnRawDataCallbackC]
        lib.set_on_snc_ready.restype = None

        lib.set_on_imu_acc_ready.argtypes = [ComputationHandle, OnRawDataCallbackC]
        lib.set_on_imu_acc_ready.restype = None

        lib.set_on_imu_gyro_ready.argtypes = [ComputationHandle, OnRawDataCallbackC]
        lib.set_on_imu_gyro_ready.restype = None

        lib.set_on_pressure_ready.argtypes = [ComputationHandle, OnPressureCallbackC]
        lib.set_on_pressure_ready.restype = None

        lib.set_on_navigation_ready.argtypes = [ComputationHandle, ctypes.c_bool]
        lib.set_on_navigation_ready.restype = None

        lib.set_on_embedded_airtouch_ready.argtypes = [ComputationHandle, ctypes.c_bool]
        lib.set_on_embedded_airtouch_ready.restype = None

        lib.set_on_gesture_ready.argtypes = [ComputationHandle, GestureCallbackC]
        lib.set_on_gesture_ready.restype = None

        lib.set_on_airmouse_button_changed.argtypes = [ComputationHandle, OnAirMouseButtonChangedCallbackC]
        lib.set_on_airmouse_button_changed.restype = None

        lib.set_model_type.argtypes = [ComputationHandle, c_int]
        lib.set_model_type.restype = None

        lib.set_license_for_app.argtypes = [c_char_p]
        lib.set_license_for_app.restype = None

        lib.set_license.argtypes = [c_int, c_char_p]
        lib.set_license.restype = None

        lib.has_license.argtypes = [ComputationHandle, c_int]
        lib.has_license.restype = ctypes.c_bool

        lib.get_licenses_request_url.argtypes = []
        lib.get_licenses_request_url.restype = c_char_p

        lib.get_licenses_key.argtypes = [c_int]
        lib.get_licenses_key.restype = c_char_p

        lib.get_firmware_command.argtypes = [c_int32]
        lib.get_firmware_command.restype = ByteBuffer

        lib.free_firmware_buffer.argtypes = [POINTER(c_uint8)]
        lib.free_firmware_buffer.restype = None


    def is_data_needed(self, index: int) -> bool:
        return bool(self._native_lib.is_data_needed(self._handle, index))

    def handle_data(self, data: bytes, length: int) -> None:
        buf = (c_uint8 * length).from_buffer_copy(data)
        self._native_lib.handle_data(self._handle, buf, c_int32(length))

    def _create_raw_callback_wrapper(self, callback: OnRawDataCallback, rms_count: int = 0):
        def _native_cb(
            timestamp: int,
            data_ptr: POINTER(c_float),
            data_len: int,
            frequency: int,
            frequency_std: float,
            rms_ptr: POINTER(c_float),
        ):
            data = [data_ptr[i] for i in range(data_len)]
            rms = [rms_ptr[i] for i in range(rms_count)] if rms_count > 0 else []
            callback(timestamp, data, frequency, frequency_std, rms)
        return OnRawDataCallbackC(_native_cb)

    def _set_raw_callback(self, setter_func, callback: OnRawDataCallback | None, rms_count: int = 0):
        if callback is None:
            setter_func(self._handle, ctypes.cast(0, OnRawDataCallbackC))
            return None
        callback_ref = self._create_raw_callback_wrapper(callback, rms_count)
        setter_func(self._handle, callback_ref)
        return callback_ref

    def set_on_snc_ready(self, callback: OnRawDataCallback | None) -> None:
        self._snc_callback_ref = self._set_raw_callback(
            self._native_lib.set_on_snc_ready, callback, self._snc_rms_count
        )

    def set_on_imu_acc_ready(self, callback: OnRawDataCallback | None) -> None:
        self._imu_acc_callback_ref = self._set_raw_callback(
            self._native_lib.set_on_imu_acc_ready, callback, 0
        )

    def set_on_imu_gyro_ready(self, callback: OnRawDataCallback | None) -> None:
        self._imu_gyro_callback_ref = self._set_raw_callback(
            self._native_lib.set_on_imu_gyro_ready, callback, 0
        )

    def set_on_pressure_ready(self, callback: OnPressureCallback | None) -> None:
        if callback is None:
            self._native_lib.set_on_pressure_ready(self._handle, ctypes.cast(0, OnPressureCallbackC))
            return None
        def _native_cb(pressure_data: float):
            callback(pressure_data)
        callback_ref = OnPressureCallbackC(_native_cb)
        self._native_lib.set_on_pressure_ready(self._handle, callback_ref)

    def set_on_navigation_ready(self, active: bool) -> None:
        self._native_lib.set_on_navigation_ready(self._handle, ctypes.c_bool(active))

    def set_air_touch_active(self, active: bool) -> None:
        self._native_lib.set_on_embedded_airtouch_ready(self._handle, ctypes.c_bool(active))

    def set_on_gesture_ready(self, callback: OnGestureDataReceivedCallback | None) -> None:
        if callback is None:
            self._native_lib.set_on_gesture_ready(self._handle, ctypes.cast(0, GestureCallbackC))
            return None
        def _native_cb(gesture_type: int):
            callback(GestureType.from_value(gesture_type))
        callback_ref = GestureCallbackC(_native_cb)
        self._gesture_callback_ref = callback_ref
        self._native_lib.set_on_gesture_ready(self._handle, callback_ref)

    def set_on_button_changed(self, callback: OnAirMouseButtonChangedReceivedCallback | None) -> None:
        if callback is None:
            self._native_lib.set_on_airmouse_button_changed(self._handle, ctypes.cast(0, OnAirMouseButtonChangedCallbackC))
            return None
        def _native_cb(button: int):
            callback(AirMouseButton.from_value(button))
        callback_ref = OnAirMouseButtonChangedCallbackC(_native_cb)
        self._airmouse_button_callback_ref = callback_ref
        self._native_lib.set_on_airmouse_button_changed(self._handle, callback_ref)

    @staticmethod
    def set_license_for_app(license_name: str):
        ComputationWrapper._ensure_native_lib_loaded()
        ComputationWrapper._initialize_native_lib()
        ComputationWrapper._native_lib.set_license_for_app(license_name.encode())

    @staticmethod
    def set_license(license: int, license_str: str):
        ComputationWrapper._ensure_native_lib_loaded()
        ComputationWrapper._initialize_native_lib()
        ComputationWrapper._native_lib.set_license(c_int(license), license_str.encode())

    def has_license(self, license: int) -> bool:
        return bool(self._native_lib.has_license(self._handle, c_int(license)))

    @staticmethod
    def get_licenses_request_url() -> str:
        ComputationWrapper._ensure_native_lib_loaded()
        ComputationWrapper._initialize_native_lib()
        result = ComputationWrapper._native_lib.get_licenses_request_url()
        # c_char_p should automatically convert to bytes, but handle both cases
        if isinstance(result, int):
            # If ctypes returns an int (pointer address), convert it to a string
            return ctypes.string_at(result).decode('utf-8')
        return result.decode('utf-8') if isinstance(result, bytes) else str(result)

    @staticmethod
    def get_licenses_key(license: int) -> str:
        ComputationWrapper._ensure_native_lib_loaded()
        ComputationWrapper._initialize_native_lib()
        result = ComputationWrapper._native_lib.get_licenses_key(c_int(license))
        # c_char_p should automatically convert to bytes, but handle both cases
        if isinstance(result, int):
            # If ctypes returns an int (pointer address), convert it to a string
            return ctypes.string_at(result).decode('utf-8')
        return result.decode('utf-8') if isinstance(result, bytes) else str(result)

    def set_model_type(self, model_type: ModelType):
        self._native_lib.set_model_type(self._handle, c_int(model_type.value_int))

    @staticmethod
    def get_firmware_command_bytes(command: int) -> bytes:
        ComputationWrapper._ensure_native_lib_loaded()
        ComputationWrapper._initialize_native_lib()
        buf = ComputationWrapper._native_lib.get_firmware_command(c_int32(command))

        array_ptr = ctypes.cast(buf.data, POINTER(c_uint8 * buf.length))
        result = bytes(array_ptr.contents)
        
        ComputationWrapper._native_lib.free_firmware_buffer(buf.data)
        return result


from setuptools import setup, find_packages
from pathlib import Path

# Get the base directory
base_dir = Path(__file__).parent

# Read the README file for the long description
readme_file = base_dir / "README.md"
long_description = ""
if readme_file.exists():
    with open(readme_file, "r", encoding="utf-8") as f:
        long_description = f.read()

# Find all library files
libs_dir = base_dir / "mudra_sdk" / "libs"
package_data = []

# Recursively find all .dll, .so, and .dylib files
if libs_dir.exists():
    for lib_file in libs_dir.rglob("*.dll"):
        rel_path = lib_file.relative_to(base_dir / "mudra_sdk")
        package_data.append(str(rel_path))
    for lib_file in libs_dir.rglob("*.so"):
        rel_path = lib_file.relative_to(base_dir / "mudra_sdk")
        package_data.append(str(rel_path))
    for lib_file in libs_dir.rglob("*.dylib"):
        rel_path = lib_file.relative_to(base_dir / "mudra_sdk")
        package_data.append(str(rel_path))

setup(
    name="mudra_sdk",
    version="0.1.13",  
    packages=find_packages(exclude=["test*", "tests*"]),
    install_requires=[
        "bleak>=0.21.0",
    ],
    include_package_data=True,
    package_data={
        "mudra_sdk": package_data,
    },
    author="Foad Khoury",
    description="Python SDK for Mudra with native library support",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.7",
    project_urls={
        "Documentation": "https://wearable-devices.github.io/#welcome"
    },
    keywords="mudra sdk bluetooth ble hardware",
    classifiers=[
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Hardware",
    ],
)


"""
Mudra SDK - Python client for Mudra API
"""

__version__ = "0.1.13"
__author__ = "Foad Khoury"

# Import main classes/functions to make them available at package level

from .models import *
from .service import *

# Define what's available when someone does: from mudra_sdk import *
__all__ = [
    "Mudra",
    "MudraDevice",  
    "FirmwareCallbacks",
]